class Conference < ActiveRecord::Base
  translates :info, :title, :payment_message

  mount_uploader :cover, CoverUploader
  mount_uploader :poster, PosterUploader

  belongs_to :conference_type
  belongs_to :city

  has_many :conference_host_organizations, dependent: :destroy
  has_many :organizations, through: :conference_host_organizations
  has_many :conference_administrators, dependent: :destroy
  has_many :administrators, through: :conference_administrators, source: :user
  has_many :event_locations
  
  has_many :workshops

  accepts_nested_attributes_for :conference_host_organizations, reject_if: proc {|u| u[:organization_id].blank?}, allow_destroy: true

    before_create :make_slug

  def to_param
    slug
  end

  def host_organization?(org)
    return false unless org.present?
    org_id = org.is_a?(Organization) ? org.id : org

    organizations.each do |o|
      return true if o.id = org_id
    end

    return false
  end

  def host?(user)
    if user.present?
      return true if user.administrator?
      
      conference_administrators.each do |u|
        return true if user.id == u.id
      end
      
      organizations.each do |o|
        return true if o.host?(user)
      end
    end
    return false
  end

  def url(action = :show)
    path(action)
  end

  def path(action = :show)
    action = action.to_sym
    '/conferences/' + conference_type.slug + '/' + slug + (action == :show ? '' : '/' + action.to_s)
  end

  def location
    return nil unless organizations.present?
    organizations.first.location
  end

  def registered?(user)
    registration = ConferenceRegistration.find_by(:user_id => user.id, :conference_id => id)
    return registration ? registration.is_attending : false
  end

  def registration_exists?(user)
    ConferenceRegistration.find_by(:user_id => user.id, :conference_id => id).present?
  end

  def registration_open
    registration_status == :open
  end

  def registration_status
    s = read_attribute(:registration_status)
    s.present? ? s.to_sym : nil
  end

  def registration_status=(new_registration_status)
    write_attribute :registration_status, new_registration_status.to_s
  end

  def make_slug(reset = false)
    if reset
      self.slug = nil
    end

    self.slug ||= Conference.generate_slug(
        conferencetype || :annual,
        conference_year,
        city_name.gsub(/\s/, '')
      )
  end

  def city_name
    return city.city if city.present?
    return location.present? ? location.city : nil
  end

  def conference_year
    self.year || (end_date.present? ? end_date.year : nil)
  end

  def over?
    return false unless end_date.present?
    return end_date < DateTime.now
  end

  def self.default_payment_amounts
    [25, 50, 100]
  end

  def self.conference_types
    {
      annual: '%{city}%{year}',
      n:      'North%{year}',
      s:      'South%{year}',
      e:      'East%{year}',
      w:      'West%{year}',
      ne:     'Northeast%{year}',
      nw:     'Northwest%{year}',
      se:     'Southeast%{year}',
      sw:     'Southwest%{year}'
    }
  end

  def self.generate_slug(type, year, city)
    Conference.conference_types[(type || :annual).to_sym].gsub('%{city}', city).gsub('%{year}', year.to_s)
  end

end
